import os
from abc import ABC, abstractmethod
from typing import Dict

from pysqlscribe.ast.base import Node
from pysqlscribe.ast.nodes import (
    FromNode,
    JoinNode,
    WhereNode,
    GroupByNode,
    OrderByNode,
    LimitNode,
    UnionNode,
    ExceptNode,
    IntersectNode,
    HavingNode,
    OffsetNode,
    SelectNode,
    InsertNode,
    ReturningNode,
)
from pysqlscribe.env_utils import str2bool
from pysqlscribe.exceptions import DialectValidationError
from pysqlscribe.regex_patterns import (
    ALIAS_SPLIT_REGEX,
    VALID_IDENTIFIER_REGEX,
    AGGREGATE_IDENTIFIER_REGEX,
    SCALAR_IDENTIFIER_REGEX,
    EXPRESSION_IDENTIFIER_REGEX,
    ALIAS_REGEX,
)
from pysqlscribe.renderers.base import Renderer


class Dialect(ABC):
    __escape_identifiers_enabled: bool = True

    def __init__(self):
        self._renderer = self.make_renderer()

    @abstractmethod
    def make_renderer(self) -> Renderer: ...

    def validate(self, current_node: Node, next_node: Node):
        valid_next = self.valid_node_transitions.get(type(current_node), ())
        if type(next_node) not in valid_next:
            raise DialectValidationError(
                f"{type(next_node).__name__} cannot follow {type(current_node).__name__}"
            )

    @property
    def valid_node_transitions(self) -> dict[type[Node], tuple[type[Node], ...]]:
        return {
            SelectNode: (FromNode,),
            FromNode: (
                JoinNode,
                WhereNode,
                GroupByNode,
                OrderByNode,
                LimitNode,
                UnionNode,
                ExceptNode,
                IntersectNode,
                OffsetNode,
            ),
            JoinNode: (WhereNode, GroupByNode, OrderByNode, LimitNode, JoinNode),
            WhereNode: (WhereNode, GroupByNode, OrderByNode, LimitNode),
            OrderByNode: (LimitNode,),
            LimitNode: (OffsetNode,),
            GroupByNode: (HavingNode, OrderByNode, LimitNode),
            HavingNode: (OrderByNode, LimitNode),
            UnionNode: (),
            ExceptNode: (),
            IntersectNode: (),
            InsertNode: (ReturningNode,),
        }

    def escape_identifier(self, identifier: str):
        if not self.escape_identifiers_enabled:
            return identifier
        return self._escape_identifier(identifier)

    @abstractmethod
    def _escape_identifier(self, identifier: str): ...

    def normalize_identifiers_args(self, *args) -> str:
        arg = args[0]
        if isinstance(arg, str):
            arg = [arg]
        identifiers = []

        for identifier in arg:
            identifier = str(identifier).strip()

            if len(parts := ALIAS_SPLIT_REGEX.split(identifier, maxsplit=1)) == 2:
                base, alias = parts[0].strip(), parts[1].strip()

                identifier = self.validate_identifier(base)
                if not ALIAS_REGEX.match(alias):
                    raise ValueError(f"Invalid SQL alias: {alias}")

                identifiers.append(f"{identifier} AS {alias}")
            else:
                identifiers.append(self.validate_identifier(identifier))

        return ", ".join(identifiers)

    def validate_identifier(self, identifier: str) -> str:
        if VALID_IDENTIFIER_REGEX.match(identifier):
            identifier = self.escape_identifier(identifier)
        elif (
            AGGREGATE_IDENTIFIER_REGEX.match(identifier)
            or SCALAR_IDENTIFIER_REGEX.match(identifier)
            or EXPRESSION_IDENTIFIER_REGEX.match(identifier)
        ):
            identifier = identifier
        else:
            raise ValueError(f"Invalid SQL identifier: {identifier}")
        return identifier

    @property
    def escape_identifiers_enabled(self):
        if not str2bool(os.environ.get("PYSQLSCRIBE_ESCAPE_IDENTIFIERS", "true")):
            return False
        return self.__escape_identifiers_enabled

    @escape_identifiers_enabled.setter
    def escape_identifiers_enabled(self, value: bool):
        self.__escape_identifiers_enabled = value

    def render(self, node: Node) -> str:
        return self._renderer.render(node)


class DialectRegistry:
    dialects: Dict[str, type[Dialect]] = {}

    @classmethod
    def register(cls, key: str):
        def decorator(dialect_class: type[Dialect]) -> type[Dialect]:
            cls.dialects[key] = dialect_class
            return dialect_class

        return decorator

    @classmethod
    def get_dialect(cls, key: str) -> Dialect:
        return cls.dialects[key]()



class InvalidNodeException(Exception): ...


class DialectValidationError(Exception): ...


class InvalidColumnNameException(Exception): ...


class InvalidColumnsException(Exception): ...


class InvalidTableNameException(Exception): ...


class InvalidSchemaNameException(Exception): ...


class InvalidJoinException(Exception): ...



from pysqlscribe.dialects.base import Dialect, DialectRegistry
from pysqlscribe.renderers.mysql import Renderer, MySQLRenderer


@DialectRegistry.register("mysql")
class MySQLDialect(Dialect):
    def make_renderer(self) -> Renderer:
        return MySQLRenderer(self)

    def _escape_identifier(self, identifier: str) -> str:
        return f"`{identifier}`"



from pysqlscribe.dialects.base import Dialect, DialectRegistry
from pysqlscribe.renderers.base import Renderer
from pysqlscribe.renderers.sqlite import SqliteRenderer


@DialectRegistry.register("sqlite")
class SQLiteDialect(Dialect):
    def make_renderer(self) -> Renderer:
        return SqliteRenderer(self)

    def _escape_identifier(self, identifier: str) -> str:
        return f'"{identifier}"'



from pysqlscribe.ast.nodes import LimitNode, OffsetNode
from pysqlscribe.renderers.base import Renderer, OFFSET

FETCH_NEXT = "FETCH NEXT"


class OracleRenderer(Renderer):
    def render_limit(self, node: LimitNode) -> str:
        return f"{FETCH_NEXT} {node.state['limit']} ROWS ONLY"

    def render_offset(self, node: OffsetNode) -> str:
        return f"{OFFSET} {node.state['offset']} ROWS"



from pysqlscribe.regex_patterns import CONSTRAINT_PREFIX_REGEX, CREATE_TABLE_REGEX
import re


def parse_create_tables(sql_text: str) -> dict[str, dict[str, list[str] | str]]:
    tables = {}

    table_regex = re.compile(
        CREATE_TABLE_REGEX,
        re.IGNORECASE | re.DOTALL,
    )

    for match in table_regex.finditer(sql_text):
        schema = match.group(2)
        table_name = match.group(3)
        columns_section = match.group(4)
        columns = []

        col_defs = re.split(r",(?![^(]*\))", columns_section)

        for col_def in col_defs:
            col_def = col_def.strip()
            if re.match(CONSTRAINT_PREFIX_REGEX, col_def, re.IGNORECASE):
                continue

            parts = col_def.split()
            if parts:
                col_name = parts[0].strip('`[]"')
                columns.append(col_name)

        tables[table_name] = {
            "schema": schema,
            "columns": columns,
        }

    return tables



def str2bool(v: str) -> bool | None:
    if v is None or isinstance(v, bool):
        return v
    if v.lower() in ("true", "1"):
        return True
    elif v.lower() in ("false", "0"):
        return False
    else:
        return False



from pysqlscribe.ast.nodes import OffsetNode, LimitNode, OrderByNode
from pysqlscribe.dialects.base import Dialect, DialectRegistry
from pysqlscribe.renderers.base import Renderer
from pysqlscribe.renderers.oracle import OracleRenderer


@DialectRegistry.register("oracle")
class OracleDialect(Dialect):
    def make_renderer(self) -> Renderer:
        return OracleRenderer(self)

    @property
    def valid_node_transitions(self):
        transitions = dict(super().valid_node_transitions)
        transitions[OrderByNode] = (OffsetNode,)
        transitions[OffsetNode] = (LimitNode,)
        transitions[LimitNode] = tuple()
        return transitions

    def _escape_identifier(self, identifier: str) -> str:
        return f'"{identifier}"'



from enum import Enum


class AggregateFunctions(str, Enum):
    MAX = "MAX"
    MIN = "MIN"
    AVG = "AVG"
    COUNT = "COUNT"
    SUM = "SUM"
    DISTINCT = "DISTINCT"

    def __str__(self):
        return self.value


class ScalarFunctions(str, Enum):
    ABS = "ABS"
    FLOOR = "FLOOR"
    CEIL = "CEIl"
    SQRT = "SQRT"
    ROUND = "ROUND"
    SIGN = "SIGN"
    LENGTH = "LENGTH"
    UPPER = "UPPER"
    LOWER = "LOWER"
    CONCAT = "CONCAT"
    POWER = "POWER"
    LN = "LN"
    EXP = "EXP"
    TRUNC = "TRUNC"
    REVERSE = "REVERSE"
    LTRIM = "LTRIM"
    RTRIM = "RTRIM"
    TRIM = "TRIM"
    NULLIF = "NULLIF"
    COALESCE = "COALESCE"
    ACOS = "ACOS"
    ASIN = "ASIN"
    ATAN = "ATAN"
    ATAN2 = "ATAN2"
    COS = "COS"
    SIN = "SIN"
    TAN = "TAN"
    SINH = "SINH"
    COSH = "COSH"
    TANH = "TANH"

    def __str__(self):
        return self.value



from pysqlscribe.column import Column, ExpressionColumn
from pysqlscribe.functions import ScalarFunctions


def _scalar_function(scalar_function: str, column: Column | str | int) -> Column | str:
    if not isinstance(column, Column):
        return f"{scalar_function}({column})"
    return ExpressionColumn(f"{scalar_function}({column})", column.table_name)


def abs_(column: Column | str):
    return _scalar_function(ScalarFunctions.ABS, column)


def floor(column: Column | str):
    return _scalar_function(ScalarFunctions.FLOOR, column)


def ceil(column: Column | str):
    return _scalar_function(ScalarFunctions.CEIL, column)


def sqrt(column: Column | str):
    return _scalar_function(ScalarFunctions.SQRT, column)


def sign(column: Column | str):
    return _scalar_function(ScalarFunctions.SIGN, column)


def length(column: Column | str):
    return _scalar_function(ScalarFunctions.LENGTH, column)


def upper(column: Column | str):
    return _scalar_function(ScalarFunctions.UPPER, column)


def lower(column: Column | str):
    return _scalar_function(ScalarFunctions.LOWER, column)


def ltrim(column: Column | str):
    return _scalar_function(ScalarFunctions.LTRIM, column)


def rtrim(column: Column | str):
    return _scalar_function(ScalarFunctions.RTRIM, column)


def trim(column: Column | str):
    return _scalar_function(ScalarFunctions.TRIM, column)


def reverse(column: Column | str):
    return _scalar_function(ScalarFunctions.REVERSE, column)


def round_(column: Column | str, decimals: int | None = None):
    if not decimals:
        return _scalar_function(ScalarFunctions.ROUND, column)
    if not isinstance(column, Column):
        return f"{ScalarFunctions.ROUND}({column}, {decimals})"
    return ExpressionColumn(
        f"{ScalarFunctions.ROUND}({column}, {decimals})", column.table_name
    )


def trunc(column: Column | str, decimals: int | None = None):
    if not decimals:
        return _scalar_function(ScalarFunctions.TRUNC, column)
    if not isinstance(column, Column):
        return f"{ScalarFunctions.TRUNC}({column}, {decimals})"
    return ExpressionColumn(
        f"{ScalarFunctions.TRUNC}({column}, {decimals})", column.table_name
    )


def power(base: Column | str | int, exponent: Column | str | int):
    if all(isinstance(arg, Column) for arg in (base, exponent)):
        return ExpressionColumn(
            f"{ScalarFunctions.POWER}({base}, {exponent})",
            base.table_name,
        )
    return f"{ScalarFunctions.POWER}({base}, {exponent})"


def ln(column: Column | str | int):
    return _scalar_function(ScalarFunctions.LN, column)


def exp(column: Column | str | int):
    return _scalar_function(ScalarFunctions.EXP, column)


def concat(*args: Column | str | int):
    if all(isinstance(arg, Column) for arg in args):
        return ExpressionColumn(
            f"{ScalarFunctions.CONCAT}({', '.join(arg.name for arg in args)})",
            args[0].table_name,
        )
    args = [f"'{arg}'" if not isinstance(arg, Column) else str(arg) for arg in args]
    return f"{ScalarFunctions.CONCAT}({', '.join(args)})"


def nullif(value1: Column | str | int, value2: Column | str | int):
    if all(isinstance(arg, Column) for arg in (value1, value2)):
        return ExpressionColumn(
            f"{ScalarFunctions.NULLIF}({value1}, {value2})",
            value1.table_name,
        )
    return f"{ScalarFunctions.NULLIF}({value1}, {value2})"


def coalesce(*args: Column | str | int):
    if all(isinstance(arg, Column) for arg in args):
        return ExpressionColumn(
            f"{ScalarFunctions.COALESCE}({', '.join(arg.name for arg in args)})",
            args[0].table_name,
        )
    args = [f"'{arg}'" if not isinstance(arg, Column) else str(arg) for arg in args]
    return f"{ScalarFunctions.COALESCE}({', '.join(args)})"


def acos(column: Column | str | int):
    return _scalar_function(ScalarFunctions.ACOS, column)


def asin(column: Column | str | int):
    return _scalar_function(ScalarFunctions.ASIN, column)


def atan(column: Column | str | int):
    return _scalar_function(ScalarFunctions.ATAN, column)


def atan2(y: Column | str | int, x: Column | str | int):
    if all(isinstance(arg, Column) for arg in (y, x)):
        return ExpressionColumn(
            f"{ScalarFunctions.ATAN2}({y}, {x})",
            y.table_name,
        )
    return f"{ScalarFunctions.ATAN2}({y}, {x})"


def cos(column: Column | str | int):
    return _scalar_function(ScalarFunctions.COS, column)


def sin(column: Column | str | int):
    return _scalar_function(ScalarFunctions.SIN, column)


def tan(column: Column | str | int):
    return _scalar_function(ScalarFunctions.TAN, column)



from typing import runtime_checkable, Protocol, Any


@runtime_checkable
class DialectProtocol(Protocol):
    def validate(self, current_node, next_node) -> None: ...

    def escape_identifier(self, identifier: str) -> str: ...
    def validate_identifier(self, identifier: str) -> str: ...
    def normalize_identifiers_args(self, args: Any) -> str: ...



from abc import ABC
from typing import Self, Any

from pysqlscribe.exceptions import InvalidNodeException, DialectValidationError
from pysqlscribe.protocols import DialectProtocol


class Node(ABC):
    next_: Self | None = None
    prev_: Self | None = None
    state: dict[str, Any]

    def __init__(self, state):
        self.state = state

    def add(self, next_: Self, dialect: DialectProtocol) -> None:
        try:
            dialect.validate(self, next_)
        except DialectValidationError as e:
            raise InvalidNodeException(f"{type(dialect).__name__}: {e}") from e
        next_.prev_ = self
        self.next_ = next_



import os
from typing import Union
from pysqlscribe.table import Table
from pysqlscribe.utils.ddl_parser import (
    parse_create_tables,
)


class InvalidPathException(Exception):
    """Custom exception for cases where a path not containing '.sql' files is provided"""


def create_tables_from_parsed(
    parsed: dict[str, dict[str, list[str] | str]], dialect: str
) -> dict[str, Table]:
    TableClass = Table.create(dialect)
    tables = {}
    for table_name, table_metadata in parsed.items():
        tables[table_name] = TableClass(
            table_name, *table_metadata["columns"], schema=table_metadata.get("schema")
        )
    return tables


def load_tables_from_ddls(
    path: Union[str, os.PathLike], dialect: str
) -> dict[str, Table]:
    all_tables = {}

    if os.path.isdir(path):
        sql_files = [
            os.path.join(path, f)
            for f in os.listdir(path)
            if f.lower().endswith(".sql")
        ]
    elif os.path.isfile(path) and path.lower().endswith(".sql"):
        sql_files = [path]
    else:
        raise InvalidPathException(
            f"Invalid path: {path}. Must be a .sql file or directory containing .sql files."
        )

    for sql_file in sql_files:
        with open(sql_file, "r") as f:
            sql_text = f.read()
        parsed = parse_create_tables(sql_text)
        table_objs = create_tables_from_parsed(parsed, dialect)
        all_tables.update(table_objs)

    return all_tables



import re

from pysqlscribe.functions import ScalarFunctions, AggregateFunctions

VALID_IDENTIFIER_REGEX = re.compile(
    r"^[A-Za-z_][A-Za-z0-9_]*(\.[A-Za-z_][A-Za-z0-9_]*)?$"
)

AGGREGATE_IDENTIFIER_REGEX = re.compile(
    rf"^({'|'.join(AggregateFunctions)})\((\*|\d+|[\w]+)\)$", re.IGNORECASE
)

# Safely build the function-name alternation
FUNC_ALT = "|".join(map(re.escape, ScalarFunctions))

# table, table.col, schema.table.col
COLUMN = r"[A-Za-z_][A-Za-z0-9_]*(?:\.[A-Za-z_][A-Za-z0-9_]*)*"

# integers, floats, .5, 1., scientific notation, with optional sign
NUMBER = r"[+\-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+\-]?\d+)?"

# single-quoted string with doubled-quote escaping: 'it''s fine'
STRING = r"'(?:[^']|'')*'"

# allow *, numbers, identifiers, or strings
ARG = rf"(?:\*|{NUMBER}|{COLUMN}|{STRING})"

SCALAR_IDENTIFIER_REGEX = re.compile(
    rf"^\s*(?:{FUNC_ALT})\s*\(\s*{ARG}(?:\s*,\s*{ARG})*\s*\)\s*$",
    re.IGNORECASE,
)


COLUMN_IDENTIFIER_REGEX = r"[A-Za-z_][A-Za-z0-9_]*(?:\.[A-Za-z_][A-Za-z0-9_]*)*"
NUMBER_REGEX = r"[+\-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+\-]?\d+)?"
TERM_REGEX = rf"(?:{COLUMN_IDENTIFIER_REGEX}|{NUMBER_REGEX})"


EXPRESSION_IDENTIFIER_REGEX = re.compile(
    rf"^\s*{TERM_REGEX}(?:\s*[\+\-\*/]\s*{TERM_REGEX})*\s*$"
)

CONSTRAINT_PREFIX_REGEX = r"^(PRIMARY|FOREIGN|CONSTRAINT|UNIQUE|INDEX)"

ALIAS_REGEX = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")

ALIAS_SPLIT_REGEX = re.compile(r"\s+AS\s+", re.IGNORECASE)

WILDCARD_REGEX = re.compile(r"^\*$")

CREATE_TABLE_REGEX = (
    r"CREATE\s+TABLE\s+(?:IF\s+NOT\s+EXISTS\s+)?((\w+)\.)?(\w+)\s*\((.*?)\);"
)



from abc import ABC
from typing import List, Self

from pysqlscribe.alias import AliasMixin
from pysqlscribe.column import Column
from pysqlscribe.exceptions import InvalidTableNameException
from pysqlscribe.query import QueryRegistry, JoinType
from pysqlscribe.regex_patterns import (
    VALID_IDENTIFIER_REGEX,
)


class Table(ABC, AliasMixin):
    __cache: dict[str, type["Table"]] = {}

    def __init__(self, name: str, *columns, schema: str | None = None):
        self.table_name = name
        self.schema = schema
        self.columns = columns

    @classmethod
    def create(cls, dialect: str):
        if dialect not in QueryRegistry.builders:
            raise ValueError(f"Unsupported dialect: {dialect}")

        class_name = f"{dialect.capitalize()}Table"

        if class_name in cls.__cache:
            return cls.__cache[class_name]

        query_class = QueryRegistry.get_builder(dialect).__class__

        class DynamicTable(query_class, Table):
            def __init__(self, name: str, *fields, schema: str | None = None):
                query_class.__init__(self)
                Table.__init__(self, name, *fields, schema=schema)

            def select(self, *columns):
                columns = [
                    f"{column.name}{column.alias}"
                    if isinstance(column, Column)
                    else column
                    for column in columns
                ]
                table_name = f"{self.table_name}{self.alias}"
                return super().select(*columns).from_(table_name)

            def order_by(self, *columns):
                columns = [
                    column.name if isinstance(column, Column) else column
                    for column in columns
                ]
                return super().order_by(*columns)

            def group_by(self, *columns):
                columns = [
                    column.name if isinstance(column, Column) else column
                    for column in columns
                ]
                return super().group_by(*columns)

            def join(
                self,
                table: str | Table,
                join_type: str = JoinType.INNER,
                condition: str | None = None,
            ) -> Self:
                if isinstance(table, Table):
                    table = f"{table.table_name}{table.alias}"
                return super().join(table, join_type, condition)

            def insert(self, *columns, **kwargs) -> Self:
                """
                Override the insert method to ensure it uses the correct table name.
                """
                columns = [
                    column.name if isinstance(column, Column) else column
                    for column in columns
                ]
                return super().insert(
                    *columns, into=self.table_name, values=kwargs.get("values")
                )

            def __repr__(self):
                return f"{self.__class__.__name__}(name={self.table_name}, columns={self.columns})"

        # Set a meaningful class name
        DynamicTable.__name__ = class_name
        cls.__cache[class_name] = DynamicTable

        return DynamicTable

    @property
    def table_name(self):
        if self.schema:
            return f"{self.schema}.{self._table_name}"
        return self._table_name

    @table_name.setter
    def table_name(self, table_name_: str):
        if not VALID_IDENTIFIER_REGEX.match(table_name_):
            raise InvalidTableNameException(f"Invalid table name {table_name_}")
        self._table_name = table_name_
        if getattr(self, "_columns", None):
            # ensure that we re-assign our column attributes with the correct fully qualified name
            self.columns = self.columns

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, columns_: List[str]):
        if getattr(self, "_columns", None):
            for column in self.columns:
                delattr(self, column)
        self._columns = columns_
        for column_name in columns_:
            setattr(
                self, column_name, Column(column_name, self._alias or self.table_name)
            )

    def as_(self, alias: str) -> Self:
        super().as_(alias)
        # ensure that we re-assign our column attributes with the correct fully qualified name (including
        # the alias instead of the full table name)
        self.columns = self.columns
        return self


MySQLTable = Table.create("mysql")
OracleTable = Table.create("oracle")
PostgresTable = Table.create("postgres")
SqliteTable = Table.create("sqlite")



import os
from typing import List

from pysqlscribe.exceptions import InvalidSchemaNameException
from pysqlscribe.regex_patterns import VALID_IDENTIFIER_REGEX
from pysqlscribe.table import Table


class Schema:
    def __init__(
        self,
        name: str,
        tables: List[Table | str] | None = None,
        dialect: str | None = None,
    ):
        self.name = name
        self.dialect = dialect
        self.tables = tables

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, schema_name: str):
        if not VALID_IDENTIFIER_REGEX.match(schema_name):
            raise InvalidSchemaNameException(f"Invalid schema name {schema_name}")
        self._name = schema_name

    @property
    def tables(self):
        return self._tables

    @tables.setter
    def tables(self, tables_: list[str | Table]):
        if all(isinstance(table, str) for table in tables_):
            tables_ = [Table.create(self.dialect)(table_name) for table_name in tables_]
        for table in tables_:
            setattr(self, table.table_name, table)
            table.schema = self.name
        self._tables = tables_

    @property
    def dialect(self):
        return self._dialect

    @dialect.setter
    def dialect(self, dialect_: str | None):
        if dialect_:
            self._dialect = dialect_
        elif env_set_dialect := os.environ.get("PYSQLSCRIBE_BUILDER_DIALECT"):
            self._dialect = env_set_dialect
        else:
            # the user may have provided no `dialect` - this is fine if they're directly supplying `Table` objects.
            self._dialect = dialect_



from enum import Enum


class JoinType(str, Enum):
    INNER: str = "INNER"
    OUTER: str = "OUTER"
    LEFT: str = "LEFT"
    RIGHT: str = "RIGHT"
    CROSS: str = "CROSS"
    NATURAL: str = "NATURAL"

    def __str__(self):
        return self.value



from typing import Self


AS = "AS"


class AliasMixin:
    _alias: str | None = None

    def as_(self, alias: str) -> Self:
        self._alias = alias
        return self

    @property
    def alias(self) -> str:
        if self._alias:
            return f" {AS} {self._alias}"
        return ""



from abc import ABC
from typing import Dict, Self

from pysqlscribe.ast.base import Node
from pysqlscribe.ast.joins import JoinType
from pysqlscribe.ast.nodes import (
    SelectNode,
    FromNode,
    JoinNode,
    GroupByNode,
    OffsetNode,
    IntersectNode,
    ExceptNode,
    HavingNode,
    OrderByNode,
    WhereNode,
    InsertNode,
    ReturningNode,
    UnionNode,
    LimitNode,
)
from pysqlscribe.dialects import (
    Dialect,
)
from pysqlscribe.dialects.base import DialectRegistry


class Query(ABC):
    node: Node | None = None
    _dialect_key: str

    def __init__(self):
        self._dialect = DialectRegistry.get_dialect(self._dialect_key)

    @property
    def dialect(self) -> Dialect:
        return self._dialect

    def select(self, *args) -> Self:
        if not self.node:
            self.node = SelectNode({"columns": list(args)})
        return self

    def from_(self, *args) -> Self:
        self.node.add(
            FromNode({"tables": list(args)}),
            self.dialect,
        )
        self.node = self.node.next_
        return self

    def insert(self, *columns, **kwargs) -> Self:
        table = kwargs.get("into")
        values = kwargs.get("values")
        if table is None or values is None:
            raise ValueError("Insert queries require `into` and `values` keywords.")
        if not self.node:
            self.node = InsertNode(
                {"columns": columns, "table": table, "values": values}
            )
        return self

    def returning(self, *args) -> Self:
        self.node.add(ReturningNode({"columns": list(args)}), self.dialect)
        self.node = self.node.next_
        return self

    def join(
        self, table: str, join_type: str = JoinType.INNER, condition: str | None = None
    ) -> Self:
        self.node.add(
            JoinNode(
                {
                    "join_type": join_type.upper(),
                    "table": table,
                    "condition": condition,
                }
            ),
            self.dialect,
        )
        self.node = self.node.next_
        return self

    def inner_join(self, table: str, condition: str):
        return self.join(table, JoinType.INNER, condition)

    def outer_join(self, table: str, condition: str):
        return self.join(table, JoinType.OUTER, condition)

    def left_join(self, table: str, condition: str):
        return self.join(table, JoinType.LEFT, condition)

    def right_join(self, table: str, condition: str):
        return self.join(table, JoinType.RIGHT, condition)

    def cross_join(self, table: str):
        return self.join(table, JoinType.CROSS)

    def natural_join(self, table: str):
        return self.join(table, JoinType.NATURAL)

    def where(self, *args) -> Self:
        where_node = WhereNode({"conditions": list(args)})
        self.node.add(where_node, self.dialect)
        self.node = self.node.next_
        return self

    def order_by(self, *args) -> Self:
        self.node.add(
            OrderByNode({"columns": list(args)}),
            self.dialect,
        )
        self.node = self.node.next_
        return self

    def limit(self, n: int | str):
        self.node.add(LimitNode({"limit": int(n)}), self.dialect)
        self.node = self.node.next_
        return self

    def offset(self, n: int | str):
        self.node.add(OffsetNode({"offset": int(n)}), self.dialect)
        self.node = self.node.next_
        return self

    def group_by(self, *args) -> Self:
        self.node.add(
            GroupByNode({"columns": list(args)}),
            self.dialect,
        )
        self.node = self.node.next_
        return self

    def having(self, *args) -> Self:
        having_node = HavingNode({"conditions": list(args)})
        self.node.add(having_node, self.dialect)
        self.node = self.node.next_
        return self

    def union(self, query: Self | str, all_: bool = False) -> Self:
        self.node.add(UnionNode({"query": query, "all": all_}), self.dialect)
        self.node = self.node.next_
        return self

    def except_(self, query: Self | str, all_: bool = False) -> Self:
        self.node.add(ExceptNode({"query": query, "all": all_}), self.dialect)
        self.node = self.node.next_
        return self

    def intersect(self, query: Self | str, all_: bool = False) -> Self:
        self.node.add(IntersectNode({"query": query, "all": all_}), self.dialect)
        self.node = self.node.next_
        return self

    def build(self, clear: bool = True) -> str:
        query = self.dialect.render(self.node)
        if clear:
            self.node = None
        return query.strip()

    def __str__(self):
        return self.build(clear=False)

    def disable_escape_identifiers(self):
        self.dialect.escape_identifiers_enabled = False
        return self

    def enable_escape_identifiers(self):
        self.dialect.escape_identifiers_enabled = True
        return self


class QueryRegistry:
    builders: Dict[str, type[Query]] = {}

    @classmethod
    def register(cls, key: str):
        def decorator(builder_class: type[Query]) -> type[Query]:
            cls.builders[key] = builder_class
            return builder_class

        return decorator

    @classmethod
    def get_builder(cls, key: str) -> Query:
        return cls.builders[key]()


@QueryRegistry.register("mysql")
class MySQLQuery(Query):
    _dialect_key = "mysql"


@QueryRegistry.register("oracle")
class OracleQuery(Query):
    _dialect_key = "oracle"


@QueryRegistry.register("postgres")
class PostgreSQLQuery(Query):
    _dialect_key = "postgres"


@QueryRegistry.register("sqlite")
class SQLiteQuery(Query):
    _dialect_key = "sqlite"



from pysqlscribe.dialects.base import Dialect
from pysqlscribe.dialects.postgres import PostgreSQLDialect
from pysqlscribe.dialects.mysql import MySQLDialect
from pysqlscribe.dialects.oracle import OracleDialect
from pysqlscribe.dialects.sqlite import SQLiteDialect


__all__ = [
    "Dialect",
    "PostgreSQLDialect",
    "MySQLDialect",
    "OracleDialect",
    "SQLiteDialect",
]



from typing import Dict, Callable

from pysqlscribe.ast.base import Node
from pysqlscribe.ast.joins import JoinType
from pysqlscribe.ast.nodes import (
    FromNode,
    JoinNode,
    WhereNode,
    GroupByNode,
    OrderByNode,
    LimitNode,
    UnionNode,
    ExceptNode,
    IntersectNode,
    HavingNode,
    OffsetNode,
    SelectNode,
    InsertNode,
    ReturningNode,
)
from pysqlscribe.protocols import DialectProtocol
from pysqlscribe.regex_patterns import WILDCARD_REGEX

SELECT = "SELECT"
FROM = "FROM"
WHERE = "WHERE"
LIMIT = "LIMIT"
JOIN = "JOIN"
ORDER_BY = "ORDER BY"
OFFSET = "OFFSET"
GROUP_BY = "GROUP BY"
HAVING = "HAVING"
ALL = "ALL"
UNION = "UNION"
UNION_ALL = f"UNION {ALL}"
EXCEPT = "EXCEPT"
EXCEPT_ALL = f"EXCEPT {ALL}"
INTERSECT = "INTERSECT"
INTERSECT_ALL = f"INTERSECT {ALL}"
INSERT = "INSERT"
INTO = "INTO"
VALUES = "VALUES"
RETURNING = "RETURNING"
AND = "AND"


class Renderer:
    def __init__(self, dialect: DialectProtocol):
        self.dialect = dialect

    @property
    def dispatch(self) -> Dict[type[Node], Callable[[Node], str]]:
        return {
            SelectNode: self.render_select,
            FromNode: self.render_from,
            WhereNode: self.render_where,
            GroupByNode: self.render_group_by,
            OrderByNode: self.render_order_by,
            LimitNode: self.render_limit,
            JoinNode: self.render_join,
            HavingNode: self.render_having,
            OffsetNode: self.render_offset,
            UnionNode: self.render_union,
            ExceptNode: self.render_except,
            IntersectNode: self.render_intersect,
            InsertNode: self.render_insert,
            ReturningNode: self.render_returning,
        }

    def render(self, node: Node) -> str:
        query = ""
        while True:
            render_method = self.dispatch.get(type(node))
            query = render_method(node) + " " + query
            node = node.prev_
            if node is None:
                break
        return query.strip()

    def render_select(self, node: SelectNode) -> str:
        columns = self._resolve_columns(*node.state["columns"])
        return f"{SELECT} {columns}"

    def render_from(self, node: FromNode) -> str:
        tables = self.dialect.normalize_identifiers_args(node.state["tables"])
        return f"{FROM} {tables}"

    def render_where(self, node: WhereNode) -> str:
        conditions = f"{f' {AND} '.join(map(str, node.state['conditions']))}"
        return f"{WHERE} {conditions}"

    def render_group_by(self, node: GroupByNode) -> str:
        columns = self.dialect.normalize_identifiers_args(node.state["columns"])
        return f"{GROUP_BY} {columns}"

    def render_order_by(self, node: OrderByNode) -> str:
        columns = self.dialect.normalize_identifiers_args(node.state["columns"])
        return f"{ORDER_BY} {columns}"

    def render_limit(self, node: LimitNode) -> str:
        return f"{LIMIT} {node.state['limit']}"

    def render_join(self, node: JoinNode) -> str:
        table = self.dialect.normalize_identifiers_args(node.table)
        return f"{node.join_type} {JOIN} {table} " + (
            f"ON {node.condition}"
            if node.join_type not in (JoinType.NATURAL, JoinType.CROSS)
            else ""
        )

    def render_having(self, node: HavingNode) -> str:
        conditions = f"{f' {AND} '.join(map(str, node.state['conditions']))}"
        return f"{HAVING} {conditions}"

    def render_offset(self, node: OffsetNode) -> str:
        return f"{OFFSET} {node.state['offset']}"

    def render_union(self, node: UnionNode) -> str:
        operation = UNION_ALL if node.state.get("all", False) else UNION
        return f"{operation} {node.query}"

    def render_except(self, node: ExceptNode) -> str:
        operation = EXCEPT_ALL if node.state.get("all", False) else EXCEPT
        return f"{operation} {node.query}"

    def render_intersect(self, node: IntersectNode) -> str:
        operation = INTERSECT_ALL if node.state.get("all", False) else INTERSECT
        return f"{operation} {node.query}"

    def render_insert(self, node: InsertNode) -> str:
        columns = self.dialect.normalize_identifiers_args(node.state["columns"])
        table = self.dialect.normalize_identifiers_args(node.state["table"])
        values = self._resolve_insert_values(
            node.state["columns"], node.state["values"]
        )
        columns = f" ({columns})" if columns else ""
        return f"{INSERT} {INTO} {table}{columns} {VALUES} {values}"

    def render_returning(self, node: ReturningNode) -> str:
        columns = self._resolve_columns(*node.state["columns"])
        return f"{RETURNING} {columns}"

    def _resolve_columns(self, *args) -> str:
        if not args:
            args = ["*"]
        if WILDCARD_REGEX.match(args[0]):
            columns = args[0]
        else:
            columns = self.dialect.normalize_identifiers_args(args)
        return columns

    @staticmethod
    def _resolve_insert_values(columns, values) -> str:
        if isinstance(values, tuple):
            values = [values]
        assert all(
            (len(columns) == 0 or len(columns) == len(value) for value in values)
        ), "Number of columns and values must match"
        values = [f"{','.join(map(str, value))}" for value in values]
        values = ",".join([f"({v})" for v in values])
        return values



from typing import Self, Iterable, Protocol, runtime_checkable

from pysqlscribe.alias import AliasMixin
from pysqlscribe.exceptions import InvalidColumnNameException
from pysqlscribe.regex_patterns import (
    VALID_IDENTIFIER_REGEX,
    AGGREGATE_IDENTIFIER_REGEX,
    SCALAR_IDENTIFIER_REGEX,
    EXPRESSION_IDENTIFIER_REGEX,
)


class Expression:
    def __init__(self, left: str, operator: str, right: str):
        self.left = left
        self.operator = operator
        self.right = right

    def __str__(self):
        return f"{self.left} {self.operator} {self.right}"

    def __repr__(self):
        return f"Expression({self.left!r}, {self.operator!r}, {self.right!r})"


@runtime_checkable
class Subqueryish(Protocol):
    @property
    def select(self) -> Self: ...


class Column(AliasMixin):
    def __init__(self, name: str, table_name: str):
        self.name = name
        self.table_name = table_name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, column_name: str):
        if not (
            VALID_IDENTIFIER_REGEX.match(column_name)
            or AGGREGATE_IDENTIFIER_REGEX.match(column_name)
            or SCALAR_IDENTIFIER_REGEX.match(column_name)
            or EXPRESSION_IDENTIFIER_REGEX.match(column_name)
        ):
            raise InvalidColumnNameException(f"Invalid column name {column_name}")
        self._name = column_name

    @property
    def fully_qualified_name(self):
        return f"{self.table_name}.{self.name}"

    def _comparison_expression(self, operator: str, other: Self | str | int):
        if isinstance(other, Column):
            return Expression(
                self.fully_qualified_name, operator, other.fully_qualified_name
            )
        elif isinstance(other, str):
            return Expression(self.fully_qualified_name, operator, f"'{other}'")
        elif isinstance(other, (int, float)):
            return Expression(self.fully_qualified_name, operator, str(other))
        raise NotImplementedError(
            "Columns can only be compared to other columns or fixed string values"
        )

    def _arithmetic_expression(self, operator: str, other: Self | str | int):
        if isinstance(other, Column):
            return ExpressionColumn(
                f"{self.fully_qualified_name} {operator} {other.fully_qualified_name}",
                self.table_name,
            )
        else:
            return ExpressionColumn(
                f"{self.fully_qualified_name} {operator} {other}", self.table_name
            )

    def _membership_expression(
        self, operator: str, other: Iterable[str | int | float] | Subqueryish
    ):
        if isinstance(other, Subqueryish):
            return Expression(self.fully_qualified_name, operator, f"({other})")
        other_list = list(other)
        if not other_list:
            raise NotImplementedError(
                "membership expressions must be created with a non-empty iterable or a subquery"
            )
        if all(isinstance(item, str) for item in other_list):
            right_side = ", ".join(f"'{item}'" for item in other_list)
            return Expression(self.fully_qualified_name, operator, f"({right_side})")
        elif all(isinstance(item, (int, float)) for item in other_list):
            right_side = ", ".join(str(item) for item in other_list)
            return Expression(self.fully_qualified_name, operator, f"({right_side})")
        else:
            raise NotImplementedError(
                "membership expressions must be created with an iterable containing items of the same type (all strings or all numbers); mixed types are not allowed"
            )

    def __str__(self):
        return self.name

    def __eq__(self, other: Self | str):
        return self._comparison_expression("=", other)

    def __lt__(self, other):
        return self._comparison_expression("<", other)

    def __gt__(self, other):
        return self._comparison_expression(">", other)

    def __le__(self, other):
        return self._comparison_expression("<=", other)

    def __ge__(self, other):
        return self._comparison_expression(">=", other)

    def __ne__(self, other):
        return self._comparison_expression("<>", other)

    def __add__(self, other):
        return self._arithmetic_expression("+", other)

    def __sub__(self, other):
        return self._arithmetic_expression("-", other)

    def __mul__(self, other):
        return self._arithmetic_expression("*", other)

    def __truediv__(self, other):
        return self._arithmetic_expression("/", other)

    def in_(self, values: Iterable[str | int | float] | Subqueryish) -> Expression:
        return self._membership_expression("IN", values)

    def not_in(self, values: Iterable[str | int | float] | Subqueryish) -> Expression:
        return self._membership_expression("NOT IN", values)

    def like(self, pattern: str) -> Expression:
        return self._comparison_expression("LIKE", pattern)

    def not_like(self, pattern: str) -> Expression:
        return self._comparison_expression("NOT LIKE", pattern)

    def ilike(self, pattern: str) -> Expression:
        return self._comparison_expression("ILIKE", pattern)

    def _between(self, low, high, operator) -> Expression:
        def resolve_column_names(low):
            if isinstance(low, Column):
                return low.fully_qualified_name
            elif isinstance(low, str):
                return f"'{low}'"
            elif isinstance(low, (int, float)):
                return str(low)

        low_expr = resolve_column_names(low)
        high_expr = resolve_column_names(high)
        return Expression(
            self.fully_qualified_name, operator, f"{low_expr} AND {high_expr}"
        )

    def between(self, low, high) -> Expression:
        return self._between(low, high, "BETWEEN")

    def not_between(self, low, high) -> Expression:
        return self._between(low, high, "NOT BETWEEN")


class ExpressionColumn(Column):
    """Representation of a column that is the result of an arithmetic operation. Main benefit is to ensure the
    fully qualified name doesn't prepend the table name each time."""

    @property
    def fully_qualified_name(self):
        return self.name



from abc import ABC

from pysqlscribe.ast.base import Node
from pysqlscribe.ast.joins import JoinType
from pysqlscribe.exceptions import InvalidJoinException


class SelectNode(Node): ...


class FromNode(Node): ...


class JoinNode(Node):
    def __init__(self, state):
        super().__init__(state)
        self.join_type = state.get("join_type", JoinType.INNER)
        self.table = state["table"]
        if (condition := state.get("condition")) and self.join_type in (
            JoinType.NATURAL,
            JoinType.CROSS,
        ):
            raise InvalidJoinException(
                "Conditions need to be supplied for any join which is not NATURAL or CROSS"
            )
        self.condition = condition


class WhereNode(Node): ...


class OrderByNode(Node): ...


class LimitNode(Node): ...


class OffsetNode(Node): ...


class GroupByNode(Node): ...


class HavingNode(Node): ...


class CombineNode(Node, ABC):
    def __init__(self, state):
        super().__init__(state)
        self.query = state["query"]
        self.all = state.get("all", False)


class UnionNode(CombineNode): ...


class ExceptNode(CombineNode): ...


class IntersectNode(CombineNode): ...


class InsertNode(Node): ...


class ReturningNode(Node): ...
